#include "lh5dec.h"
/*  Extract lh5 data */

void fwrite_txt(lh5_decode_t *lh5, unsigned char  *p, int n)
{
	while (--n >= 0)
		if (*p != '\015' && *p != '\032') 
			*lh5->out_buf++ = *p;
}

/* ------------------------------------------------------------------------ */
unsigned short calccrc(lh5_decode_t *lh5, unsigned char  *p,unsigned int n)
{
	while (n-- > 0)
		UPDATE_CRC(*p++);
	return lh5->crc;
}


void fwrite_crc(lh5_decode_t *lh5, unsigned char  *p, int n)
{
	calccrc(lh5, p, n);

	memcpy(lh5->out_buf,p,n);
	lh5->out_buf +=n;
}


/* ------------------------------------------------------------------------ */
void fillbuf(lh5_decode_t *lh5, unsigned char n)			/* Shift bitbuf n bits left, read n bits */
{
	while (n > lh5->bitcount) {
		n -= lh5->bitcount;
		lh5->bitbuf = (lh5->bitbuf << lh5->bitcount) + (lh5->subbitbuf >> (CHAR_BIT - lh5->bitcount));
		if (lh5->compsize != 0) {
			lh5->compsize--;
			lh5->subbitbuf = *lh5->in_buf++;
		}
		else
			lh5->subbitbuf = 0;
		lh5->bitcount = CHAR_BIT;
	}
	lh5->bitcount -= n;
	lh5->bitbuf = (lh5->bitbuf << n) + (lh5->subbitbuf >> (CHAR_BIT - n));
	lh5->subbitbuf <<= n;
}

void init_getbits(lh5_decode_t *lh5)
{
	lh5->bitbuf = 0;
	lh5->subbitbuf = 0;
	lh5->bitcount = 0;
	fillbuf(lh5, 2 * CHAR_BIT);
}



/* ------------------------------------------------------------------------ */
unsigned short getbits(lh5_decode_t *lh5, unsigned char n)
{
	unsigned short  x;

	x = lh5->bitbuf >> (2 * CHAR_BIT - n);
	fillbuf(lh5, n);
	return x;
}


void make_table(lh5_decode_t *lh5, short nchar,unsigned char bitlen[],short tablebits,unsigned short table[])
{
	unsigned short  count[17];	/* count of bitlen */
	unsigned short  weight[17];	/* 0x10000ul >> bitlen */
	unsigned short  start[17];	/* first code of bitlen */
	unsigned short  total;
	unsigned int    i, l;
	int             j, k, m, n, avail;
	unsigned short *p;

	avail = nchar;

	/* initialize */
	for (i = 1; i <= 16; i++) {
		count[i] = 0;
		weight[i] = 1 << (16 - i);
	}

	/* count */
	for (i = 0; i < nchar; i++)
		count[bitlen[i]]++;

	/* calculate first code */
	total = 0;
	for (i = 1; i <= 16; i++) {
		start[i] = total;
		total += weight[i] * count[i];
	}
	if ((total & 0xffff) != 0)
		fprintf(stderr,"make_table(), Bad table (5)\n");

	/* shift data for make table. */
	m = 16 - tablebits;
	for (i = 1; i <= tablebits; i++) {
		start[i] >>= m;
		weight[i] >>= m;
	}

	/* initialize */
	j = start[tablebits + 1] >> m;
	k = 1 << tablebits;
	if (j != 0)
		for (i = j; i < k; i++)
			table[i] = 0;

	/* create table and tree */
	for (j = 0; j < nchar; j++) {
		k = bitlen[j];
		if (k == 0)
			continue;
		l = start[k] + weight[k];
		if (k <= tablebits) {
			/* code in table */
			for (i = start[k]; i < l; i++)
				table[i] = j;
		}
		else {
			/* code not in table */
			p = &table[(i = start[k]) >> m];
			i <<= tablebits;
			n = k - tablebits;
			/* make tree (n length) */
			while (--n >= 0) {
				if (*p == 0) {
					lh5->right[avail] = lh5->left[avail] = 0;
					*p = avail++;
				}
				if (i & 0x8000)
					p = &lh5->right[*p];
				else
					p = &lh5->left[*p];
				i <<= 1;
			}
			*p = j;
		}
		start[k] = l;
	}
}


/* ------------------------------------------------------------------------ */
static void read_pt_len(lh5_decode_t *lh5,short nn,short nbit,short i_special)
{
	int           i, c, n;

	n = getbits(lh5, nbit);
	if (n == 0) {
		c = getbits(lh5, nbit);
		for (i = 0; i < nn; i++)
			lh5->pt_len[i] = 0;
		for (i = 0; i < 256; i++)
			lh5->pt_table[i] = c;
	}
	else {
		i = 0;
		while (i < n) {
			c = lh5->bitbuf >> (16 - 3);
			if (c == 7) {
				unsigned short  mask = 1 << (16 - 4);
				while (mask & lh5->bitbuf) {
					mask >>= 1;
					c++;
				}
			}
			fillbuf(lh5, (c < 7) ? 3 : c - 3);
			lh5->pt_len[i++] = c;
			if (i == i_special) {
				c = getbits(lh5, 2);
				while (--c >= 0)
					lh5->pt_len[i++] = 0;
			}
		}
		while (i < nn)
			lh5->pt_len[i++] = 0;
		make_table(lh5, nn, lh5->pt_len, 8, lh5->pt_table);
	}
}

/* ------------------------------------------------------------------------ */
static void read_c_len(lh5_decode_t *lh5)
{
	short           i, c, n;

	n = getbits(lh5, CBIT);
	if (n == 0) {
		c = getbits(lh5, CBIT);
		for (i = 0; i < NC; i++)
			lh5->c_len[i] = 0;
		for (i = 0; i < 4096; i++)
			lh5->c_table[i] = c;
	} else {
		i = 0;
		while (i < n) {
			c = lh5->pt_table[lh5->bitbuf >> (16 - 8)];
			if (c >= NT) {
				unsigned short  mask = 1 << (16 - 9);
				do {
					if (lh5->bitbuf & mask)
						c = lh5->right[c];
					else
						c = lh5->left[c];
					mask >>= 1;
				} while (c >= NT);
			}
			fillbuf(lh5, lh5->pt_len[c]);
			if (c <= 2) {
				if (c == 0)
					c = 1;
				else if (c == 1)
					c = getbits(lh5, 4) + 3;
				else
					c = getbits(lh5, CBIT) + 20;
				while (--c >= 0)
					lh5->c_len[i++] = 0;
			}
			else
				lh5->c_len[i++] = c - 2;
		}
		while (i < NC)
			lh5->c_len[i++] = 0;
		make_table(lh5, NC, lh5->c_len, 12, lh5->c_table);
	}
}


/* ------------------------------------------------------------------------ */
unsigned short decode_c_st1(lh5_decode_t *lh5)
{
	unsigned short  j, mask;

	if (lh5->blocksize == 0) {
		lh5->blocksize = getbits(lh5, 16);
		read_pt_len(lh5, NT, TBIT, 3);
		read_c_len(lh5);
		read_pt_len(lh5, lh5->np, lh5->pbit, -1);
	}
	lh5->blocksize--;
	j = lh5->c_table[lh5->bitbuf >> 4];
	if (j < NC)
		fillbuf(lh5, lh5->c_len[j]);
	else {
		fillbuf(lh5, 12);
		mask = 1 << (16 - 1);
		do {
			if (lh5->bitbuf & mask)
				j = lh5->right[j];
			else
				j = lh5->left[j];
			mask >>= 1;
		} while (j >= NC);
		fillbuf(lh5, lh5->c_len[j] - 12);
	}
	return j;
}

/* ------------------------------------------------------------------------ */
unsigned short decode_p_st1(lh5_decode_t *lh5)
{
	unsigned short  j, mask;

	j = lh5->pt_table[lh5->bitbuf >> (16 - 8)];
	if (j < lh5->np)
		fillbuf(lh5, lh5->pt_len[j]);
	else {
		fillbuf(lh5, 8);
		mask = 1 << (16 - 1);
		do {
			if (lh5->bitbuf & mask)
				j = lh5->right[j];
			else
				j = lh5->left[j];
			mask >>= 1;
		} while (j >= lh5->np);
		fillbuf(lh5, lh5->pt_len[j] - 8);
	}
	if (j != 0)
		j = (1 << (j - 1)) + getbits(lh5, j - 1);
	return j;
}

/* ------------------------------------------------------------------------ */
void decode_start_st1(lh5_decode_t *lh5)
{
	lh5->np = 14;
	lh5->pbit = 4;
	
	init_getbits(lh5);
	lh5->blocksize = 0;
}

/* ------------------------------------------------------------------------ */
void lh5_decode(lh5_decode_t *lh5, unsigned char *outp, unsigned char *inp, long packed_size, long original_size)
{
	unsigned int i, j, k, c;
	unsigned int dicsiz1, offset;
	unsigned char *dtext;
	
	lh5->compsize = packed_size;
	lh5->origsize = original_size;
	lh5->in_buf = inp;
	lh5->out_buf = outp;

	lh5->crc = 0;
	lh5->dicsiz = 1L << 13 /* dicbit*/;
	dtext = (unsigned char *) malloc(lh5->dicsiz);
	if (dtext == NULL)
	{
		fprintf(stderr, "No mem\n");
		goto exit;
	}
	for (i=0; i<lh5->dicsiz; i++) dtext[i] = 0x20;
	decode_start_st1(lh5);
	dicsiz1 = lh5->dicsiz - 1;
	offset = 0x100 - 3;
	lh5->count = 0;
	lh5->loc = 0;
	while (lh5->count < lh5->origsize) {
		c = decode_c_st1(lh5);
		if (c <= UCHAR_MAX) {
			dtext[lh5->loc++] = c;
			if (lh5->loc == lh5->dicsiz) {
				fwrite_crc(lh5, dtext, lh5->dicsiz);
				lh5->loc = 0;
			}
			lh5->count++;
		}
		else {
			j = c - offset;
			i = (lh5->loc - decode_p_st1(lh5) - 1) & dicsiz1;
			lh5->count += j;
			for (k = 0; k < j; k++) {
				c = dtext[(i + k) & dicsiz1];

				dtext[lh5->loc++] = c;
				if (lh5->loc == lh5->dicsiz) {
					fwrite_crc(lh5, dtext, lh5->dicsiz);
					lh5->loc = 0;
				}
			}
		}
	}
	if (lh5->loc != 0) {
		fwrite_crc(lh5, dtext, lh5->loc);
	}

exit:
	free(dtext);
}
