#include <E32Base.h>
#include <D32snd.h> 
#include "epocsound.h"

RDevSound iDevSound;		 // Sound device driver
TRequestStatus iPlayStatus;  // iPlayStatus; 
TInt iNrBytesBeingPlayed;    // Played samples counter


const TInt KBufferSize = 8000; // buffer for one second

TInt InitSoundDevice(void)
	{

#ifdef __WINS__
TInt err=User::LoadPhysicalDevice(_L("ESDRV.PDD"));
	if (!err)
		err=User::LoadLogicalDevice(_L("ESOUND.LDD"));
	if (err && err!=KErrAlreadyExists && err!=KErrNotFound) // third clause to aid porting to other platforms
		User::Leave(err);
#endif 

	TInt ret = iDevSound.Open();
	if ( ret != KErrNone )
		return(ret);

	TSoundCaps sndCaps;
	iDevSound.Caps(sndCaps);
    ret = sndCaps().iService;
    ret = sndCaps().iVolume;
    ret = sndCaps().iMaxVolume;
    ret = sndCaps().iMaxFrequency;

	TSoundConfig sndConfig;
	iDevSound.Config(sndConfig); // read the default config

	sndConfig().iVolume = EVolumeLoud;
	sndConfig().iAlawBufferSize = KBufferSize;
	
	ret = iDevSound.SetConfig(sndConfig); // read the default config
	if(ret != KErrNone)
		return(ret);
		
	// prepare the buffer
	ret = iDevSound.PreparePlayAlawBuffer(); // This *must* be done before reading the Buffer size!!!
	if(ret != KErrNone)
		return(ret);
	iNrBytesBeingPlayed = 0; 
    return(KErrNone);

	}
TInt PlaySoundBuffer(const TUint8 *aBuffPtr, TInt aBuffLen)
{
	TPtrC8 BufferPtr(aBuffPtr,aBuffLen);
#ifdef __MARM__
	if(iNrBytesBeingPlayed)
		if(iNrBytesBeingPlayed == iDevSound.BytesPlayed())
			{
			CloseSoundDevice();
			if(InitSoundDevice())
				return(-1); // playing stalled reinitialize whole sound device
			}
#endif
	iNrBytesBeingPlayed = iDevSound.BytesPlayed();
	iDevSound.PlayAlawData(iPlayStatus, BufferPtr);
	User::WaitForRequest(iPlayStatus);
	return(iPlayStatus.Int());
}

TInt CloseSoundDevice(void)
{
  iDevSound.PlayCancel();
  iDevSound.FlushPlayBuffer();
  iDevSound.ResetBytesPlayed(); 
  iDevSound.Close();
  return(KErrNone);
}