#include <stdio.h>
#include <stdlib.h>
//#include <getopt.h>

/* Standard includes for OSS DSP using
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
*/

#include "ayemu.h"

#include "epocsound.h"

static const int DEBUG = 1;
static const int AY_LOG = /*1*/0;
static /* const */ int SAVE_TO_FILE = 0;

/*
struct option options[] = {
  { "quiet",   no_argument, NULL, 'q'},
  { "rand",    no_argument, NULL, 'Z'},
  { "stdout",  no_argument, NULL, 's'},
  { "verbose", no_argument, NULL, 'v'},
  { "usage",   no_argument, NULL, 'u'},
  { "help",    no_argument, NULL, 'h'},
  { 0, 0, 0, 0}
};
*/

ayemu_ay_t ay;
ayemu_vtx_t vtx;
unsigned char regs[14];
uint8_t * audio_buf, *alaw_buf;
int audio_bufsize,alaw_bufsize;
//int audio_fd;
FILE *audio_fd, *aylog_fd;
int freq, chans, bits;

// #define DEVICE_NAME "/dev/dsp"
// #define DEVICE_NAME "snd.raw"
#define AY_LOG_NAME "ay.log"

int WriteEPOCRecordHeader(FILE *fdst,int len)
{
 const TUint8 EPOCRecordHeader[] =
 {
  0x37, 0x00, 0x00, 0x10, /* UID1 */
  0x6d, 0x00, 0x00, 0x10, /* UID2 */
  0x7e, 0x00, 0x00, 0x10, /* UID3 */
  0xcf, 0xac, 0x08, 0x55, /* CRC UID */

  0x14, 0x00, 0x00, 0x00,
  0x04,

  0x52, 0x00, 0x00, 0x10,
  0x34, 0x00, 0x00, 0x00,
  0x89, 0x00, 0x00, 0x10,
  0x25, 0x00, 0x00, 0x00,
  0x7e, 0x00, 0x00, 0x10,

  0x2a,
  0x52, 0x65, 0x63, 0x6f, 0x72, 0x64, 0x2e, 0x61, 0x70, 0x70 /* record.app */
 }; /* unsigned char EPOCRecordHeader1[] */

#define C_ADPCMNoCompress 0x00000000L // No compression

 TUint32 UncompressedRecordDataLength = len;
 const TUint32 CompressFlag = C_ADPCMNoCompress;
 const TUint16 NumRepeat = 0;
 const TUint8 Volume = 5;
 // 1- Minimum and 5 maximum, if Cyclesample 0- add volume
 // 0- Increase volume (for cycled sample): First play at volume level 3
 //    and for all next play at volume level 5. (Example Z:\system\alarms\ChurchBell)
 const TUint8 DummyByte = 0;  // Must be 0
 const TUint32 RepeatDelay = 0;
 TUint32 RecordDataLength = UncompressedRecordDataLength;
  /* if compression is enabled then it's size of compressed block
     else same value like UncompressedRecordDataLength */
/* ------------------------------------------------------------------------------------------ */
 if ((len = fwrite(EPOCRecordHeader,sizeof(EPOCRecordHeader), 1, fdst)) != 1) 
 {
  fprintf (stderr, "Error writting to audio file, break.\n");
  return(0);
 }

 fwrite(&UncompressedRecordDataLength, sizeof(UncompressedRecordDataLength), 1, fdst);
 fwrite(&CompressFlag, sizeof(CompressFlag), 1, fdst);
 fwrite(&NumRepeat, sizeof(NumRepeat), 1, fdst);
 fwrite(&Volume, sizeof(Volume), 1, fdst);
 fwrite(&DummyByte, sizeof(DummyByte), 1, fdst);
 fwrite(&RepeatDelay, sizeof(RepeatDelay), 1, fdst);
 fwrite(&RecordDataLength, sizeof(RecordDataLength), 1, fdst);
 return(1);
}

void usage ()
{
    printf(
	   "AY/YM VTX format player\n"
	   "Symbian EPOC32 port by: Petr Petyovsky aka Poke^Patisoners 2006\n"
	   "(Build:"__DATE__","__TIME__" for JHCon visitors)\n"
	   "It uses libayemu (http://libayemu.sourceforge.net)\n"
	   "Version 0.90 (2005/02/10). Written and copyrights by Alexander Sashnov.\n\n"
);
//	   "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n\n"
//	   "Usage: playvtx [option(s)] files...\n"
//	   "  -q --quiet\tquiet (don't print title)\n"
//	   "  -Z --random\trandom play\n"
//	   "  -s --stdout\twrite to stdout\n"
//	   "  -v --verbose\tincrease verbosity level\n"
//	   "  -h --help\n"
//	   "  -u --usage\tthis help\n"
//	   );
}

/* play vtx file */
void Play (const char *filename);

int
     main (int argc, char **argv)
     {
       int index;
	   usage();
       /*
       int qflag = 0;  // quite plating
       int Zflag = 0;  // random list
       int sflag = 0;  // to stdout
       int vflag = 0;  // verbose
       
       int c;
//        `getopt_long' stores the option index here.

       int option_index = 0;
	   
       opterr = 0;
     
       while ((c = getopt_long (argc, argv, "+qZsvuh", options, &option_index)) != -1)
         switch (c)
           {
           case 'q':
             qflag = 1;
             break;
           case 'Z':
             Zflag = 1;
             break;
           case 's':
             sflag = 1;
             break;
           case 'v':
             vflag = 1;
             break;
	   case 'h':
	   case 'u':
	     usage ();
	     exit (0);
           case '?':
             if (isprint (optopt))
               fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	     else
               fprintf (stderr,
                        "Unknown option character `\\x%x'.\n",
                        optopt);
	     usage ();
             return 1;
	   case -1:
	     break;
           default:
             abort ();
           }
     
       if (DEBUG)
       printf ("qflag = %d, Zflag = %d, sflag = %d, vflag = %d\n", 
	       qflag, Zflag, sflag, vflag);
     
       if (Zflag)
	 printf ("The -Z flag is not implemented yet, sorry\n");
*/
	 index = 1;
#ifdef __MARM__
     if( (argc > 2) && (argv[1][0] == '-') && (argv[1][1] == 'f') )
		{
		SAVE_TO_FILE = 1;
		index ++;
		}
     for (/*index = optind*/; index < argc; index++)
	  {
	   if (DEBUG) 
	      printf ("Playing file %s\n", argv[index]);
	   Play (argv[index]);
	  }
#else
	 //Play ("c:\\csoon.vtx");
     //Play ("c:\\Melange.vtx");
     Play ("c:\\emergency.vtx");
     //Play ("c:\\2ndreal1.vtx");
     //Play ("c:\\refresh.vtx");
#endif
       return 0;
     }


void Play (const char *filename)
{
  int len;
  int err;
  int audio_bufsize;
  char outputfilename[64];

  sprintf(outputfilename,"%.58s.alaw",filename);
  memset (&ay, 0, sizeof(ay));

//  freq = 44100;
//  chans = 2;  /* 1=mono, 2=stereo */
  freq = 8000; /* er5 default */
  chans = 1;  /* 1=mono er5 default */
  bits = 16;  /* 16 or 8 bit */

  /* Try to load VTX file */
  if (! ayemu_vtx_open (&vtx, filename)) {
    fprintf (stderr, "Can't open vtx file %s\n", filename);
  }
  else if (! ayemu_vtx_load_data (&vtx)) {
    fprintf (stderr, "Can't unpack vtx data");
    goto free_vtx;
  }
    /* =====----- OSS Init sound system ------=======
  else if ((audio_fd = open(DEVICE_NAME, O_WRONLY, 0)) == -1) {
    fprintf (stderr, "Can't open /dev/dsp\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &bits) == -1) {
    fprintf (stderr, "Can't set sound format\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &chans) == -1) {
    fprintf (stderr, "Can't set number of channels\n");
  }
  else if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &freq) == -1) {
    fprintf (stderr, "Can't set audio freq\n");
  }
  */
  //else if ((audio_fd = fopen(DEVICE_NAME, "w")) == NULL) {
  //  fprintf (stderr, "Can't open"DEVICE_NAME"\n");
  else if (SAVE_TO_FILE && ((audio_fd = fopen(outputfilename, "wb")) == NULL)) {
    fprintf (stderr, "Can't open %s\n",outputfilename);
  }
  else if ((!SAVE_TO_FILE) && (InitSoundDevice())) {
	fprintf (stderr, "Can't open Sound Device\n");
  }
  else if (AY_LOG && ((aylog_fd = fopen(AY_LOG_NAME, "w")) == NULL)) {
    fprintf (stderr, "Can't open"AY_LOG_NAME"\n");
  }


  else {

    /* VTX song info */
    printf ("Title: %s\n", &vtx.Title);
    printf ("Author: %s\n", &vtx.Author);
    printf ("From : %s\n", &vtx.From);
    printf ("Tracker : %s\n", &vtx.Tracker);
    printf ("Comment : %s\n", &vtx.Comment);
    printf ("Chip type: %s\tStereo: %d\tLoop: %d\n", vtx.Chip, vtx.Stereo, vtx.Loop);
    printf ("Chip freq: %d\tPlayer Freq:%d\tYear: %d\n", vtx.ChipFreq, vtx.PlayerFreq, vtx.Year);

    if(SAVE_TO_FILE)
	  if (DEBUG) printf ("FILE sound system initialization success: bits=%d, chans=%d, freq=%d\n", bits, chans, freq);
	else
	  if (DEBUG) printf ("EPOC sound system initialization success: bits=%d, chans=%d, freq=%d\n", bits, chans, freq);

    /* Allocate audio buffer for one AY frame */
    audio_bufsize = freq * chans * (bits >> 3) / vtx.PlayerFreq;
    if ((audio_buf = (uint8_t *) malloc (audio_bufsize)) == NULL) {
      fprintf (stderr, "Can't allocate sound buffer\n");
      goto free_audio;
    }

    alaw_bufsize = freq * chans / vtx.PlayerFreq; // We use 8bit Alaw
    if ((alaw_buf = (uint8_t *) malloc (alaw_bufsize)) == NULL) {
      fprintf (stderr, "Can't allocate Alaw sound buffer\n");
      goto free_alaw;
    }

    // Writing EPOC Sound Record header
	if(SAVE_TO_FILE)
		WriteEPOCRecordHeader(audio_fd, alaw_bufsize * vtx.regdata_size /14); 

    ayemu_set_chip (&ay, (strcmp(vtx.Chip,"ay") ==0)? AYEMU_AY:AYEMU_YM,
        vtx.ChipFreq, (chans==1)?(AYEMU_MONO):(ayemu_stereo_t)vtx.Stereo);
    if (DEBUG) printf ("ayemu_set_chip (%2s, freq %d, stereo %d)\n",vtx.Chip, vtx.ChipFreq, (chans==1)?(AYEMU_MONO):(ayemu_stereo_t)vtx.Stereo);

    ayemu_start (&ay, freq, chans, bits);
    if (DEBUG) printf ("ayemu_start (freq %d, chans %d, bits %d)\n",freq,chans,bits);

    while (ayemu_vtx_get_next_frame (&vtx, regs)) {
      ayemu_set_regs (&ay, regs);
      if (AY_LOG)
	   fprintf (aylog_fd, "VTX pos %4d Regs: tones: %4d %4d %4d %2d R7: %d%d%d %d%d%d vols: %2d %2d %2d env: %5d %2d\n", 
        vtx.pos,
		ay.reg_tone_a, ay.reg_tone_b, ay.reg_tone_c, ay.reg_noise,
		ay.reg_R7_tone_a, ay.reg_R7_tone_b, ay.reg_R7_tone_c,
		ay.reg_R7_noise_a, ay.reg_R7_noise_b, ay.reg_R7_noise_c,
		ay.reg_vol_a, ay.reg_vol_b, ay.reg_vol_c,
		ay.reg_env_freq, ay.reg_env_style);

      ayemu_gen_sound (&ay, audio_buf, audio_bufsize);
      conv_u16bit_alaw((unsigned short *)audio_buf, alaw_buf, alaw_bufsize);
	  if(SAVE_TO_FILE)
		{
		/* save sound buffer */
        if ((len = fwrite(alaw_buf, alaw_bufsize, 1, audio_fd)) != 1) 
			{
			fprintf (stderr, "Error writting to audio file, break.\n");
			break;
			}		
		}
	  else
		{
		/* play sound buffer */
		if(err = PlaySoundBuffer(alaw_buf, alaw_bufsize))
			{
			fprintf (stderr, "Error writting to sound device (err.%d), break.\n",err);
			break;
			}		
		}
    }

  free(alaw_buf);

free_alaw:
  free(audio_buf);

free_audio:
  if (AY_LOG) printf ("Closing aylog file... \n");
    fclose (aylog_fd);
  if (AY_LOG) printf ("Done.\n");

  if(SAVE_TO_FILE)
	{
	if (DEBUG) printf ("Closing audiofile... \n");
	fclose (audio_fd);
	if (DEBUG) printf ("Done.\n");
	}
  else
	{
	if (DEBUG) printf ("Closing sound device... \n");
	CloseSoundDevice();
	if (DEBUG) printf ("Done.\n");
	}

free_ay:
  if (DEBUG) printf ("ayemu_free(&ay)... \n");
    ayemu_free(&ay);
  if (DEBUG) printf ("Done.\n");

free_vtx:
  if (DEBUG) printf ("ayemu_vtx_free(&vtx)... \n");
    ayemu_vtx_free(&vtx);
  if (DEBUG) printf ("Done.\n");
  }
}
